# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest
import time
import copy
import uuid
from concurrent.futures import ThreadPoolExecutor

from towhee.runtime.node_repr import NodeRepr
from towhee.runtime.nodes import create_node, NodeStatus
from towhee.runtime.data_queue import DataQueue, ColumnType, Empty
from towhee.runtime.operator_manager import OperatorPool


class TestFilterNode(unittest.TestCase):
    """
    Filter node test.
    """
    node_info = {
        'inputs': ('num', ),
        'outputs': ('larger_than_5', ),
        'op_info': {
            'type': 'hub',
            'operator': 'local/filter_operator',
            'tag': 'main',
            'init_args': [],
            'init_kws': {'threshold': 5}
        },
        'iter_info': {
            'type': 'filter',
            'param': {'filter_by': ['num']}
        },
        'config': {'name': 'testa'},
        'next_nodes': ['_output']
    }

    def setUp(self):
        self.node_repr = NodeRepr.from_dict(uuid.uuid4().hex, self.node_info)
        self.op_pool = OperatorPool()
        self.thread_pool = ThreadPoolExecutor()

    def test_normal(self):
        in_que = DataQueue([('url', ColumnType.SCALAR), ('num', ColumnType.QUEUE)])
        size = 10
        for i in range(size):
            in_que.put(('test_url', i))
        in_que.seal()
        out_que1 = DataQueue([('url', ColumnType.SCALAR), ('num', ColumnType.QUEUE), ('larger_than_5', ColumnType.QUEUE)])
        out_que2 = DataQueue([('larger_than_5', ColumnType.QUEUE)])
        node = create_node(self.node_repr, self.op_pool, [in_que], [out_que1, out_que2])
        node.initialize()
        f = self.thread_pool.submit(node.process)
        f.result()
        self.assertEqual(node.status, NodeStatus.FINISHED)
        self.assertTrue(out_que1.sealed)
        self.assertTrue(out_que2.sealed)
        for i in range(size):
            self.assertEqual(out_que1.get_dict(),
                            {
                                'url': 'test_url',
                                'num': i,
                                'larger_than_5': i + 6 if i < 4 else Empty()
                            })
            if i > 5:
                self.assertEqual(out_que2.get_dict(),
                                {
                                    'larger_than_5': i
                                })

    def test_multithread(self):
        in_que = DataQueue([('url', ColumnType.SCALAR), ('num', ColumnType.QUEUE)])
        out_que1 = DataQueue([('url', ColumnType.SCALAR), ('num', ColumnType.QUEUE), ('larger_than_5', ColumnType.QUEUE)])
        out_que2 = DataQueue([('larger_than_5', ColumnType.QUEUE)])
        node = create_node(self.node_repr, self.op_pool, [in_que], [out_que1, out_que2])
        self.assertTrue(node.initialize())
        f = self.thread_pool.submit(node.process)

        def write(size):
            for i in range(size):
                in_que.put(('test_url', i))
                time.sleep(0.01)
            in_que.seal()

        size = 10
        w_f = self.thread_pool.submit(write, size)
        w_f.result()
        f.result()
        self.assertTrue(node.status == NodeStatus.FINISHED)
        self.assertEqual(out_que1.size, size)
        self.assertEqual(out_que2.size, size - 6)
        for i in range(size):
            self.assertEqual(out_que1.get_dict(),
                             {
                                 'url': 'test_url',
                                 'num': i,
                                 'larger_than_5': i + 6 if i < 4 else Empty()
                             })
            if i > 5:
                self.assertEqual(out_que2.get_dict(),
                                {
                                    'larger_than_5': i
                                })
        self.assertEqual(out_que1.size, 0)
        self.assertEqual(out_que2.size, 0)
        self.assertIsNone(out_que1.get())
        self.assertIsNone(out_que2.get())

    def test_schema_cover(self):
        in_que = DataQueue([('url', ColumnType.SCALAR), ('num', ColumnType.QUEUE)])
        size = 10
        for i in range(size):
            in_que.put(('test_url', i))
        in_que.seal()
        out_que1 = DataQueue([('url', ColumnType.SCALAR), ('num', ColumnType.QUEUE)])
        out_que2 = DataQueue([('num', ColumnType.QUEUE)])
        node_info = copy.deepcopy(self.node_info)
        node_info['outputs'] = ('num', )
        node_repr = NodeRepr.from_dict('test_node', node_info)
        node = create_node(node_repr, self.op_pool, [in_que], [out_que1, out_que2])
        self.assertTrue(node.initialize())
        f = self.thread_pool.submit(node.process)
        f.result()
        self.assertTrue(node.status == NodeStatus.FINISHED)
        self.assertTrue(out_que1.sealed)
        self.assertTrue(out_que2.sealed)
        self.assertEqual(out_que1.size, 4)
        self.assertEqual(out_que2.size, 4)
        for i in range(6, size):
            self.assertEqual(out_que1.get_dict(),
                            {
                                'url': 'test_url',
                                'num': i
                            })

            self.assertEqual(out_que2.get_dict(),
                            {
                                'num': i
                            })

    def test_schema_cover2(self):
        in_que = DataQueue([('num1', ColumnType.QUEUE), ('num2', ColumnType.QUEUE)])
        size = 10
        for i in range(size):
            in_que.put((i, i + 1))
        in_que.seal()
        out_que1 = DataQueue([('num1', ColumnType.QUEUE), ('num2', ColumnType.QUEUE)])
        out_que2 = DataQueue([('num2', ColumnType.QUEUE)])
        node_info = copy.deepcopy(self.node_info)
        node_info['inputs'] = ('num1', 'num2')
        node_info['outputs'] = ('num2', 'num1')
        node_info['iter_info']['param']['filter_by'] = ['num1']
        node_repr = NodeRepr.from_dict('test_node', node_info)
        node = create_node(node_repr, self.op_pool, [in_que], [out_que1, out_que2])
        self.assertTrue(node.initialize())
        f = self.thread_pool.submit(node.process)
        f.result()
        self.assertTrue(node.status == NodeStatus.FINISHED)
        self.assertTrue(out_que1.sealed)
        self.assertTrue(out_que2.sealed)
        self.assertEqual(out_que1.size, 4)
        self.assertEqual(out_que2.size, 4)
        for i in range(6, size):
            self.assertEqual(out_que1.get_dict(),
                            {
                                'num1': i + 1,
                                'num2': i
                            })

    def test_output_with_scalar(self):
        in_que = DataQueue([('url', ColumnType.SCALAR), ('num', ColumnType.QUEUE)])
        size = 10
        for i in range(size):
            in_que.put(('test_url', i))
        in_que.seal()
        out_que1 = DataQueue([('url', ColumnType.SCALAR), ('larger_than_5', ColumnType.QUEUE)])
        node = create_node(self.node_repr, self.op_pool, [in_que], [out_que1])
        self.assertTrue(node.initialize())
        f = self.thread_pool.submit(node.process)
        f.result()
        self.assertTrue(node.status == NodeStatus.FINISHED)
        self.assertTrue(out_que1.sealed)
        for i in range(6, size):
            self.assertEqual(out_que1.get_dict(),
                            {
                                'url': 'test_url',
                                'larger_than_5': i
                            })

    def test_output_all_scalar(self):
        in_que = DataQueue([('url', ColumnType.SCALAR), ('num', ColumnType.QUEUE)])
        size = 10
        for i in range(size):
            in_que.put(('test_url', i))
        in_que.seal()
        out_que1 = DataQueue([('url', ColumnType.SCALAR)])
        node = create_node(self.node_repr, self.op_pool, [in_que], [out_que1])
        self.assertTrue(node.initialize())
        f = self.thread_pool.submit(node.process)
        f.result()
        self.assertTrue(node.status == NodeStatus.FINISHED)
        self.assertTrue(out_que1.sealed)
        self.assertTrue(out_que1.size == 1)
        for i in range(size):
            if i == 0:
                self.assertEqual(out_que1.get_dict(),
                                {
                                    'url': 'test_url'
                                })
            else:
                self.assertEqual(out_que1.get_dict(), None)

    def test_stopped(self):
        in_que = DataQueue([('url', ColumnType.SCALAR), ('num', ColumnType.QUEUE)])
        in_que.put(('test_url', 1))
        in_que.put(('test_url', 1))
        in_que.seal()
        out_que1 = DataQueue([('url', ColumnType.SCALAR), ('num', ColumnType.QUEUE), ('larger_than_5', ColumnType.QUEUE)])
        out_que1.seal()
        node = create_node(self.node_repr, self.op_pool, [in_que], [out_que1])
        self.assertTrue(node.initialize())
        f = self.thread_pool.submit(node.process)
        f.result()
        self.assertEqual(node.status, NodeStatus.STOPPED)
        self.assertTrue(out_que1.sealed)
        self.assertTrue(out_que1.size == 0)

    def test_failed(self):
        in_que = DataQueue([('url', ColumnType.SCALAR), ('num', ColumnType.QUEUE)])
        in_que.put(('test_url', 'a'))
        in_que.seal()
        out_que1 = DataQueue([('url', ColumnType.SCALAR), ('num', ColumnType.QUEUE), ('vec', ColumnType.QUEUE)])
        out_que2 = DataQueue([('vec', ColumnType.QUEUE)])
        node = create_node(self.node_repr, self.op_pool, [in_que], [out_que1, out_que2])
        self.assertTrue(node.initialize())
        f = self.thread_pool.submit(node.process)
        f.result()
        self.assertTrue(node.status == NodeStatus.FAILED)
        self.assertTrue(out_que1.sealed)
        self.assertTrue(out_que2.sealed)
        self.assertTrue(out_que1.size == 0)
        self.assertTrue(out_que2.size == 0)

    def test_all_empty(self):
        node_info = {
            'inputs': ('num', 'another'),
            'outputs': ('vec', 'vec2'),
            'op_info': {
                'type': 'lambda',
                'operator': lambda x: x < 2,
                'tag': 'main',
                'init_args': None,
                'init_kws': {}
            },
            'iter_info': {
                'type': 'filter',
                'param': {'filter_by': ['num']}
            },
            'config': {'name': 'testa'},
            'next_nodes': ['_output']
        }
        node_repr = NodeRepr.from_dict('test_node', node_info)
        in_que = DataQueue([('url', ColumnType.SCALAR), ('num', ColumnType.QUEUE), ('another', ColumnType.QUEUE), ('some', ColumnType.QUEUE)])
        in_que.put(('test_url', 1, 1, 1))
        in_que.put(('test_url', Empty(), Empty(), 1))
        in_que.seal()
        out_que = DataQueue(
            [
                ('url', ColumnType.SCALAR), ('num', ColumnType.QUEUE), ('another', ColumnType.QUEUE), ('some', ColumnType.QUEUE),
                ('vec', ColumnType.QUEUE)
            ]
        )
        node = create_node(node_repr, self.op_pool, [in_que], [out_que])
        self.assertTrue(node.initialize())
        f = self.thread_pool.submit(node.process)
        f.result()
        self.assertEqual(out_que.get_dict(),
                            {
                                'url': 'test_url',
                                'num': 1,
                                'another': 1,
                                'some': 1,
                                'vec': 1,
                            })
        self.assertEqual(out_que.get_dict(),
                            {
                                'url': 'test_url',
                                'num': Empty(),
                                'another': Empty(),
                                'some': 1,
                                'vec': Empty(),
                            })

    def test_scalar_empty(self):
        node_info = {
            'inputs': ('url', 'num'),
            'outputs': ('vec2', 'vec'),
            'op_info': {
                'type': 'lambda',
                'operator': lambda x: x < 2,
                'tag': 'main',
                'init_args': None,
                'init_kws': {}
            },
            'iter_info': {
                'type': 'filter',
                'param': {'filter_by': ['num']}
            },
            'config': {'name': 'testa'},
            'next_nodes': ['_output']
        }
        node_repr = NodeRepr.from_dict('test_node', node_info)
        in_que = DataQueue([('url', ColumnType.SCALAR), ('num', ColumnType.QUEUE), ('another', ColumnType.QUEUE)])
        in_que.put(('test_url', 1, 1))
        in_que.put(('test_url', Empty(), 1))
        in_que.seal()
        out_que = DataQueue([('url', ColumnType.SCALAR), ('num', ColumnType.QUEUE), ('another', ColumnType.QUEUE), ('vec', ColumnType.QUEUE)])
        node = create_node(node_repr, self.op_pool, [in_que], [out_que])
        self.assertTrue(node.initialize())
        f = self.thread_pool.submit(node.process)
        f.result()
        self.assertEqual(out_que.get_dict(),
                            {
                                'url': 'test_url',
                                'num': 1,
                                'another': 1,
                                'vec': 1,
                            })
        self.assertEqual(out_que.get_dict(),
                            {
                                'url': 'test_url',
                                'num': Empty(),
                                'another': 1,
                                'vec': Empty(),
                            })

    def test_queue_empty(self):
        node_info = {
            'inputs': ('num', 'another'),
            'outputs': ('vec', 'vec2'),
            'op_info': {
                'type': 'lambda',
                'operator': lambda x, y: x < 2 and y < 2,
                'tag': 'main',
                'init_args': None,
                'init_kws': {}
            },
            'iter_info': {
                'type': 'filter',
                'param': {'filter_by': ['num', 'another']}
            },
            'config': {'name': 'test'},
            'next_nodes': ['_output']
        }
        node_repr = NodeRepr.from_dict('test_node', node_info)
        in_que = DataQueue([('url', ColumnType.SCALAR), ('num', ColumnType.QUEUE), ('another', ColumnType.QUEUE)])
        in_que.put(('test_url', 1, 1))
        in_que.put(('test_url', Empty(), 1))
        in_que.seal()
        out_que = DataQueue(
            [
                ('url', ColumnType.SCALAR), ('num', ColumnType.QUEUE), ('another', ColumnType.QUEUE), ('vec', ColumnType.QUEUE),
                ('vec2', ColumnType.QUEUE)
            ]
        )
        node = create_node(node_repr, self.op_pool, [in_que], [out_que])
        self.assertTrue(node.initialize())
        f = self.thread_pool.submit(node.process)
        f.result()
        self.assertEqual(out_que.get_dict(),
                            {
                                'url': 'test_url',
                                'num': 1,
                                'another': 1,
                                'vec': 1,
                                'vec2': 1
                            })
        self.assertEqual(out_que.get_dict(),
                            {
                                'url': 'test_url',
                                'num': Empty(),
                                'another': 1,
                                'vec': Empty(),
                                'vec2': Empty()
                            })

    def test_empty_without_next(self):
        node_info = {
            'inputs': ('num', 'another'),
            'outputs': ('vec', 'vec2'),
            'op_info': {
                'type': 'lambda',
                'operator': lambda x: x < 2,
                'tag': 'main',
                'init_args': None,
                'init_kws': {}
            },
            'iter_info': {
                'type': 'filter',
                'param': {'filter_by': ['num']}
            },
            'config': {'name': 'testa'},
            'next_nodes': None
        }
        node_repr = NodeRepr.from_dict('test_node', node_info)
        in_que = DataQueue([('url', ColumnType.SCALAR), ('num', ColumnType.QUEUE), ('another', ColumnType.QUEUE), ('some', ColumnType.QUEUE)])
        in_que.put(('test_url', 1, 1, 1))
        in_que.put(('test_url', Empty(), 1, 1))
        in_que.seal()
        out_que = DataQueue(
            [
                ('url', ColumnType.SCALAR), ('num', ColumnType.QUEUE), ('another', ColumnType.QUEUE), ('some', ColumnType.QUEUE),
                ('vec', ColumnType.QUEUE)
            ]
        )
        node = create_node(node_repr, self.op_pool, [in_que], [out_que])
        self.assertTrue(node.initialize())
        f = self.thread_pool.submit(node.process)
        f.result()
        self.assertEqual(node.status, NodeStatus.FINISHED)
        self.assertEqual(out_que.get_dict(),
                            {
                                'url': 'test_url',
                                'num': 1,
                                'another': 1,
                                'some': 1,
                                'vec': 1,
                            })
        self.assertEqual(out_que.get_dict(),
                            {
                                'url': 'test_url',
                                'num': Empty(),
                                'another': 1,
                                'some': 1,
                                'vec': Empty(),
                            })


if __name__ == '__main__':
    unittest.main()
